import datetime
import operator

from cloud_governance.common.clouds.aws.cloudtrail.cloudtrail_operations import CloudTrailOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy
from operator import ge


class EC2Stop(NonClusterZombiePolicy):
    """
    This class delete the stopped ec2 instance
    Send mail to the user if the ec2 instance is stopped > 20, 25th days
    Delete the instance when it stopped 30 days
    """

    FIRST_MAIL_NOTIFICATION_INSTANCE_DAYS = 20
    SECOND_MAIL_NOTIFICATION_INSTANCE_DAYS = 25
    DELETE_INSTANCE_DAYS = 30

    def __init__(self):
        super().__init__()
        self._cloudtrail = CloudTrailOperations(region_name=self._region)

    def run(self):
        """
        This method list all stopped instances for more than 30 days and terminate if dry_run no
        @return:
        """
        return self.__fetch_stop_instance(sign=ge, instance_days=self.FIRST_MAIL_NOTIFICATION_INSTANCE_DAYS, delete_instance_days=self.DELETE_INSTANCE_DAYS)

    def __fetch_stop_instance(self, instance_days: int, delete_instance_days: int, sign: operator = ge):
        """
        This method list all stopped instances for more than 30 days and terminate if dry_run no
        @return:
        """
        instances = self._ec2_client.describe_instances(Filters=[{'Name': 'instance-state-name', 'Values': ['stopped']}])['Reservations']
        stopped_instances = []
        stopped_instance_tags = {}
        ec2_types = {}
        stopped_time = ''
        days = 0
        for instance in instances:
            for resource in instance['Instances']:
                instance_id = resource.get('InstanceId')
                stopped_time = self._cloudtrail.get_stop_time(resource_id=instance_id, event_name='StopInstances')
                if not stopped_time:
                    stopped_time = datetime.datetime.now()
                days = self._calculate_days(create_date=stopped_time)
                if days in (instance_days, self.SECOND_MAIL_NOTIFICATION_INSTANCE_DAYS):
                    user = self._get_tag_name_from_tags(tags=resource.get('Tags'), tag_name='User')
                    if user:
                        self.__trigger_mail(tags=resource.get('Tags'), stopped_time=stopped_time, resource_id=instance_id, days=days, ec2_type=resource.get("InstanceType"), instance_id=instance_id)
                    else:
                        logger.info('User is missing')
                if sign(days, instance_days):
                    if days >= delete_instance_days:
                        stopped_instance_tags[instance_id] = resource.get('Tags')
                        ec2_types[instance_id] = resource.get('InstanceType')
                    stopped_instances.append([resource.get('InstanceId'), self._get_tag_name_from_tags(tags=resource.get('Tags'), tag_name='Name'),
                                              self._get_tag_name_from_tags(tags=resource.get('Tags'), tag_name='User'), str(resource.get('LaunchTime')),
                                              self._get_tag_name_from_tags(tags=resource.get('Tags'), tag_name='Policy')
                                              ])
        if self._dry_run == "no":
            for instance_id, tags in stopped_instance_tags.items():
                if self._get_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                    tags.append({'Key': 'Policy', 'Value': 'backup'})
                    tag_specifications = [{'ResourceType': 'image', 'Tags': tags}]
                    if sign == ge:
                        tag_specifications.append({'ResourceType': 'snapshot', 'Tags': tags})
                    try:
                        ami_id = self._ec2_client.create_image(InstanceId=instance_id, Name=self._get_tag_name_from_tags(tags=tags), TagSpecifications=tag_specifications)['ImageId']
                        self.__trigger_mail(tags=tags, stopped_time=stopped_time, days=days, resource_id=instance_id, image_id=ami_id, ec2_type=ec2_types[instance_id], instance_id=instance_id)
                        self._ec2_client.terminate_instances(InstanceIds=[instance_id])
                        logger.info(f'Deleted the instance: {instance_id}')
                    except Exception as err:
                        logger.info(err)
        return stopped_instances

    def __trigger_mail(self, tags: list, stopped_time: str, resource_id: str, days: int, image_id: str = '', ec2_type: str = '', instance_id: str = ''):
        """
        This method send triggering mail
        @param tags:
        @param stopped_time:
        @param resource_id:
        @param days:
        @return:
        """
        try:
            special_user_mails = self._literal_eval(self._special_user_mails)
            user, instance_name = self._get_tag_name_from_tags(tags=tags, tag_name='User'), self._get_tag_name_from_tags(tags=tags, tag_name='Name')
            to = user if user not in special_user_mails else special_user_mails[user]
            ldap_data = self._ldap.get_user_details(user_name=to)
            cc = [self._account_admin, f'{ldap_data.get("managerId")}@redhat.com']
            subject, body = self._mail_description.ec2_stop(name=ldap_data.get('displayName'), days=days, image_id=image_id, delete_instance_days=self.DELETE_INSTANCE_DAYS, instance_name=instance_name, resource_id=resource_id, stopped_time=stopped_time, ec2_type=ec2_type)
            self._mail.send_email_postfix(to=to, content=body, subject=subject, cc=cc, resource_id=instance_id)
        except Exception as err:
            logger.info(err)



import os
import datetime
from ast import literal_eval

import boto3

from cloud_governance.common.clouds.aws.cloudtrail.cloudtrail_operations import CloudTrailOperations
from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.policy.aws.zombie_cluster_resource import ZombieClusterResources


class NonClusterZombiePolicy:

    DAYS_TO_DELETE_RESOURCE = 7
    DAYS_TO_TRIGGER_RESOURCE_MAIL = 4

    def __init__(self):
        self._end_date = datetime.datetime.now()
        self._start_date = self._end_date - datetime.timedelta(days=self.DAYS_TO_DELETE_RESOURCE)
        self._account = os.environ.get('account', '')
        self._dry_run = os.environ.get('dry_run', 'yes')
        self._region = os.environ.get('AWS_DEFAULT_REGION', 'us-east-2')
        self._policy = os.environ.get('policy', '')
        self._policy_output = os.environ.get('policy_output', '')
        self._ec2_client = boto3.client('ec2', region_name=self._region)
        self._ec2_operations = EC2Operations(region=self._region)
        self._s3_client = boto3.client('s3')
        self._iam_operations = IAMOperations()
        self._iam_client = boto3.client('iam')
        self._cluster_prefix = 'kubernetes.io/cluster'
        self._zombie_cluster = ZombieClusterResources(cluster_prefix=self._cluster_prefix)
        self._s3operations = S3Operations(region_name=self._region)
        self._cloudtrail = CloudTrailOperations(region_name=self._region)
        self._special_user_mails = os.environ.get('special_user_mails', '{}')
        self._account_admin = os.environ.get('account_admin', '')
        self._mail = Postfix()
        self._mail_description = MailMessage()
        self.__ldap_host_name = os.environ.get('LDAP_HOST_NAME', '')
        self._ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)
        self._active_clusters = self._zombie_cluster.all_cluster_instance()

    def _literal_eval(self, data: any):
        tags = {}
        if data:
            tags = literal_eval(data)
        return tags

    def _check_live_cluster_tag(self, tags: list):
        """
        This method returns True if it is live cluster tag is active False not
        @param tags:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key').startswith(self._cluster_prefix):
                    return tag.get('Key') in self._active_clusters.values()
        return False

    def _get_tag_name_from_tags(self, tags: list, tag_name: str = 'Name'):
        """
        This method returns the tag_name from resource tags
        @param tags:
        @param tag_name:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key').strip().lower() == tag_name.lower():
                    return tag.get('Value').strip()
            return ''
        return ''

    def _calculate_days(self, create_date: datetime):
        """
        This method return the days
        @return:
        """
        today = datetime.date.today()
        days = today - create_date.date()
        return days.days

    def _get_ami_ids(self):
        """
        This method returns all image ids
        @return:
        """
        images = self._ec2_operations.get_images()
        image_ids = []
        for image in images:
            image_ids.append(image.get('ImageId'))
        return image_ids

    def _beautify_upload_data(self, upload_resource_data: list):
        """
        This method beautify the data to upload to elasticsearch
        @param upload_resource_data:
        @return:
        """
        upload_data = []
        for resource_data in upload_resource_data:
            if isinstance(resource_data, list):
                resource_str = ''
                for resource in resource_data[:-1]:
                    resource_str += f'{resource} | '
                resource_str += resource_data[-1]
                upload_data.append(resource_str)
            else:
                upload_data.append(resource_data)
        return upload_data

    def _get_resource_username(self, resource_id: str, resource_type: str, create_date: datetime = '', event_type: str = ''):
        """
        Get Username from the cloudtrail
        @param create_date:
        @param resource_id:
        @param resource_type:
        @return:
        """
        if event_type:
            return self._cloudtrail.get_username_by_instance_id_and_time(start_time=create_date, resource_id=resource_id,
                                                                         resource_type=resource_type, event_type=event_type)
        else:
            return self._cloudtrail.get_username_by_instance_id_and_time(start_time=create_date, resource_id=resource_id,
                                                                         resource_type=resource_type)

    def _get_policy_value(self, tags: list):
        """
        This method beautify the value
        @param tags:
        @return:
        """
        policy_value = self._get_tag_name_from_tags(tags=tags, tag_name='Policy').strip()
        if policy_value:
            return policy_value.replace('_', '').replace('-', '').upper()
        return 'NA'

    def _trigger_mail(self, tags: list, resource_id: str, days: int, resource_type: str):
        """
        This method send triggering mail
        @param tags:
        @param resource_id:
        @return:
        """
        try:
            special_user_mails = self._literal_eval(self._special_user_mails)
            user, resource_name = self._get_tag_name_from_tags(tags=tags, tag_name='User'), self._get_tag_name_from_tags(
                tags=tags, tag_name='Name')
            if not resource_name:
                resource_name = self._get_tag_name_from_tags(tags=tags, tag_name='cg-Name')
            to = user if user not in special_user_mails else special_user_mails[user]
            ldap_data = self._ldap.get_user_details(user_name=to)
            cc = [self._account_admin, f'{ldap_data.get("managerId")}@redhat.com']
            name = to
            if ldap_data:
                name = ldap_data.get('displayName')
            subject, body = self._mail_description.resource_message(name=name, days=days,
                                                                    notification_days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL,
                                                                    delete_days=self.DAYS_TO_DELETE_RESOURCE,
                                                                    resource_name=resource_name, resource_id=resource_id,
                                                                    resource_type=resource_type)
            self._mail.send_email_postfix(to=to, content=body, subject=subject, cc=cc, resource_id=resource_id)
        except Exception as err:
            logger.info(err)

    def _update_tag_value(self, tags: list, tag_name: str, tag_value: str):
        """
        This method updates the tag value
        @param tags:
        @param tag_name:
        @param tag_value:
        @return:
        """
        found = False
        if tags:
            for tag in tags:
                if tag.get('Key') == tag_name:
                    tag['Value'] = str(tag_value)
                    found = True
            if not found:
                tags.append({'Key': tag_name, 'Value': str(tag_value)})
            return tags
        return [{'Key': tag_name, 'Value': str(tag_value)}]

    def _get_resource_last_used_days(self, tags: list):
        """
        This method get last used day from the tags
        @param tags:
        @return:
        """
        last_used_day = self._get_tag_name_from_tags(tags=tags, tag_name='LastUsedDay')
        if not last_used_day:
            last_used_day = 1
        else:
            last_used_day = int(last_used_day) + 1
        return last_used_day

    def __delete_resource_on_name(self, resource_id: str):
        try:
            if self._policy == 'empty_buckets':
                self._s3_client.delete_bucket(Bucket=resource_id)
            elif self._policy == 'empty_roles':
                self._iam_client.delete_role(RoleName=resource_id)
            elif self._policy == 'ebs_unattached':
                self._ec2_client.delete_volume(VolumeId=resource_id)
            elif self._policy == 'zombie_elastic_ips':
                self._ec2_client.release_address(AllocationId=resource_id)
            elif self._policy == 'zombie_nat_gateways':
                self._ec2_client.delete_nat_gateway(NatGatewayId=resource_id)
            elif self._policy == 'zombie_snapshots':
                self._ec2_client.delete_snapshot(SnapshotId=resource_id)
            logger.info(f'{self._policy} deleted: {resource_id}')
        except Exception as err:
            logger.info(f'Exception raised: {err}: {resource_id}')

    def _check_resource_and_delete(self, resource_name: str, resource_id: str, resource_type: str, resource: dict, empty_days: int, days_to_delete_resource: int, tags: list = []):
        """
        This method check and delete resources
        @param resource_name:
        @param resource_id:
        @param resource_type:
        @param resource:
        @param empty_days:
        @param days_to_delete_resource:
        @param tags:
        @return:
        """
        resource_id = resource.get(resource_id)
        if not tags:
            tags = resource.get('Tags') if resource.get('Tags') else []
        user = self._get_tag_name_from_tags(tag_name='User', tags=tags)
        if not user:
            user = self._get_resource_username(resource_id=resource_id, resource_type=resource_type, event_type='EventName')
            if user:
                tags.append({'Key': 'User', 'Value': user})
        zombie_resource = {}
        if empty_days >= self.DAYS_TO_TRIGGER_RESOURCE_MAIL:
            if empty_days == self.DAYS_TO_TRIGGER_RESOURCE_MAIL:
                self._trigger_mail(resource_type=resource_name, resource_id=resource_id, tags=tags, days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL)
            elif empty_days >= days_to_delete_resource:
                if self._dry_run == 'no':
                    if self._get_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                        self._trigger_mail(resource_type=resource_name, resource_id=resource_id, tags=tags, days=empty_days)
                        self.__delete_resource_on_name(resource_id=resource_id)
            zombie_resource = resource
        return zombie_resource

    def _update_resource_tags(self, resource_id: str, left_out_days: int, tags: list, resource_left_out: bool):
        """
        This method update the tags in aws
        @return:
        """
        if left_out_days < 7 or self._dry_run == 'yes' or self._get_policy_value(tags=tags) in ('NOTDELETE', 'SKIP'):
            if self._get_tag_name_from_tags(tags=tags, tag_name='LastUsedDay') or resource_left_out:
                tags = self._update_tag_value(tags=tags, tag_name='LastUsedDay', tag_value=str(left_out_days))
                try:
                    if self._policy == 'empty_buckets':
                        self._s3_client.put_bucket_tagging(Bucket=resource_id, Tagging={'TagSet': tags})
                    elif self._policy == 'empty_roles':
                        self._iam_client.tag_role(RoleName=resource_id, Tags=tags)
                    elif self._policy in ('zombie_elastic_ips', 'zombie_nat_gateways', 'zombie_snapshots'):
                        self._ec2_client.create_tags(Resources=[resource_id], Tags=tags)
                except Exception as err:
                    logger.info(f'Exception raised: {err}: {resource_id}')



import boto3
from boto3.dynamodb.types import TypeSerializer, TypeDeserializer

from cloud_governance.common.logger.init_logger import logger


class DynamoDbOperations:

    def __init__(self, region_name: str = 'us-east-2'):
        self.__region_name = region_name
        self.__db_client = boto3.client('dynamodb', region_name=self.__region_name)
        self.__iam_client = boto3.client('iam')
        self.__serializer = TypeSerializer()
        self.__deserializer = TypeDeserializer()
        self.__db_resource = boto3.resource('dynamodb', region_name=self.__region_name)

    def __get_default_user_tags(self):
        try:
            user = self.__iam_client.get_user()['User']
            if user.get('Tags'):
                return user.get('Tags')
            return [{'Key': 'User', 'Value': user.get('UserName')}]
        except:
            []

    def serialize_data_dynamodb_data(self, item: dict):
        """
        This method return the dict of DynamoDb data
        @param item:
        @return:
        """
        return {key: self.__serializer.serialize(value) for key, value in item.items()}

    def deserialize_dynamodb_data(self, item: dict):
        """
        This method deserialize the dynamodb data to dict
        @param item:
        @return:
        """
        return {key: self.__deserializer.deserialize(value) for key, value in item.items()}

    def create_table(self, table_name: str, key_name: str):
        """
        This method create table
        @param key_name:
        @param table_name:
        @return:
        """
        try:
            tags = self.__get_default_user_tags()
            if tags:
                self.__db_client.create_table(TableName=table_name, Tags=tags,
                                              AttributeDefinitions=[{'AttributeName': key_name, 'AttributeType': 'S'}],
                                              KeySchema=[{'AttributeName': key_name, 'KeyType': 'HASH'}],
                                              ProvisionedThroughput={'ReadCapacityUnits': 10, 'WriteCapacityUnits': 10})
            else:
                self.__db_client.create_table(TableName=table_name,
                                              AttributeDefinitions=[{'AttributeName': key_name, 'AttributeType': 'S'}],
                                              KeySchema=[{'AttributeName': key_name, 'KeyType': 'HASH'}],
                                              ProvisionedThroughput={'ReadCapacityUnits': 10, 'WriteCapacityUnits': 10})
        except Exception as err:
            print(err)

    def put_item(self, table_name: str, item: dict):
        """
        This method put the data into the table
        @param item:
        @param table_name:
        @return:
        """
        try:
            response = self.__db_client.put_item(TableName=table_name, Item=item)
            return response
        except Exception as err:
            print(err)

    def scan_table(self, table_name: str, scan_kwargs):
        """
        This method scan the table and list the  queries
        @param table_name:
        @param scan_kwargs:
        @return:
        """
        responses = []
        try:
            table = self.__db_resource.Table(table_name)
            response = table.scan(**scan_kwargs)
            responses.extend(response['Items'])
            while response.get('LastEvaluatedKey'):
                response = table.scan(**scan_kwargs, ExclusiveStartKey=response['LastEvaluatedKey'])
                responses.extend(response['Items'])

        except Exception as err:
            responses = []
        return responses

    def get_item(self, table_name: str, key_name: str, item_type: str, item: str):
        """ This method get item """
        try:
            response = self.__db_client.get_item(TableName=table_name, Key={key_name: {item_type: item}})
            return response['Item']
        except Exception as err:
            raise

    def update_item(self, table_name: str, key_name: str, key_type: str, key_value: str, update_item: str,
                    item_type: str, update_value: str):
        try:
            response = self.__db_client.update_item(TableName=table_name, Key={key_name: {key_type: key_value}},
                                                    AttributeUpdates={update_item: {'Action': 'PUT',
                                                                                    'Value': {item_type: update_value}}})
            return response
        except Exception as err:
            raise

    # def batch_write(self, table_name: str, items: list):
    #     """
    #     This method groups put_item into the dynamoDb
    #     @return:
    #     """
    #     table_resource = self.__db_resource.Table(table_name)
    #     with table_resource.batch_writer() as writer:
    #         for item in items:
    #             writer.put_item(Item=item)



import os


class IBMOperations:
    """
    This class contain the common parameters of ibm polices
    """

    def __init__(self):
        self._account = os.environ.get('account', '')
        self._dry_run = os.environ.get('dry_run', 'yes')
        self._policy = os.environ.get('policy', '')
        self._tag_operation = os.environ.get('tag_operation', 'read')







import datetime
import gzip
import json
import os
import boto3
import typeguard
from botocore.exceptions import ClientError
from os import listdir
from os.path import isfile, join

from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class S3Operations:
    """ This class is responsible for S3 operations """

    def __init__(self, region_name,  report_file_name: str = "zombie_report.json", resource_file_name: str = "resources.json.gz"):
        #  @Todo ask AWS support regarding about this issue
        if region_name == 'eu-south-1':
            self.__s3_client = boto3.client('s3', region_name='us-east-1')
        else:
            self.__s3_client = boto3.client('s3', region_name=region_name)
        self.__region = region_name
        self.__report_file_name = report_file_name
        self.__resource_file_name = resource_file_name
        self.__report_file_full_path = os.path.join(os.path.dirname(__file__), self.__report_file_name)
        self.__resources_file_full_path = os.path.join(os.path.dirname(__file__), self.__resource_file_name)

    @logger_time_stamp
    @typeguard.typechecked
    def upload_file(self, file_name_path: str, bucket: str, key: str, upload_file: str):
        """
        This method upload file to s3
        STANDARD_IA , ONEZONE_IA , INTELLIGENT_TIERING , GLACIER , or DEEP_ARCHIVE
        :param file_name_path:'D:\\Performance\\Projects\\py-image-service\\data\\data\\DJI_0100.jpg'
        :param bucket:'devops-ais'
        :param key:'test-data'
        :param upload_file:'DJI_0100.jpg'
        :return:
        """

        try:
            self.__s3_client.upload_file(Filename=file_name_path,
                                         Bucket=bucket,
                                         Key=f'{key}/{upload_file}',
                                         ExtraArgs={'ServerSideEncryption': 'AES256',
                                                    'StorageClass': 'ONEZONE_IA'})
        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def download_file(self, bucket: str, key: str, download_file: str, file_name_path: str):
        """
        This method download file from s3
        :param bucket:'devops-ais'
        :param key:'logs/ec2-idle/2021/01/19/18'
        :param download_file: 'DJI_0100.jpg'
        :param file_name_path:'D:\\Performance\\Projects\\py-image-service\\data\\rt_results\\DJI_0100.jpg'
        :return:
        """
        try:
            if download_file:
                self.__s3_client.download_file(Bucket=bucket, Key=f'{key}/{download_file}', Filename=file_name_path)
            else:
                self.__s3_client.download_file(Bucket=bucket, Key=key, Filename=file_name_path)
        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def delete_file(self, bucket: str, key: str, file_name: str):
        """
        This method download file from s3
        :param bucket:'devops-ais'
        :param key:'test-data'
        :param file_name: 'DJI_0100.jpg'
        :return:
        """
        try:
            self.__s3_client.delete_object(Bucket=bucket, Key=f'{key}/{file_name}')
        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def delete_folder(self, bucket: str, key: str):
        """
        This method download file from s3
        :param bucket:'devops-ais'
        :param key:'framework/test'
        :return:
        """
        try:
            objects_to_delete = self.__s3_client.list_objects(Bucket=bucket, Prefix=key)
            delete_keys = {
                'Objects': [{'Key': k} for k in [obj['Key'] for obj in objects_to_delete.get('Contents', [])]]}
            if delete_keys['Objects']:
                self.__s3_client.delete_objects(Bucket=bucket, Delete=delete_keys)
        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def create_folder(self, bucket: str, key: str):
        """
        This method download file from s3
        :param bucket:'devops-ais'
        :param key:'framework/test'
        :return:
        """
        try:
            self.__s3_client.put_object(Bucket=bucket, Key=key)
        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def file_exist(self, bucket: str, key: str, file_name: str):
        """
        This method check if file exist
        :param bucket:'devops-ais'
        :param key:'framework/test'
        :param file_name:'file.txt'
        :return:
        """
        try:
            response = self.__s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
            if response.get('Contents'):
                for item in response['Contents']:
                    if file_name in item['Key']:
                        return True
            return False

        # Todo add custom error
        except ClientError:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def upload_objects(self, local_source: str, s3_target: str):
        """
        This method upload local data folder to s3 target path
        :param local_source: local data folder i.e. 'D:/Temp/'
        :param s3_target: target s3 path i.e. 'data_store/calc_image_data/'
        :return:
        """
        try:
            if '/' in s3_target:
                targets = s3_target.split('/')
                bucket = targets[0]
                key = '/'.join(targets[1:])
            else:
                bucket = s3_target
                key = ''

            files = [f for f in listdir(local_source) if isfile(join(local_source, f))]
            for file in files:
                filename = os.path.join(local_source, file)
                self.upload_file(file_name_path=filename, bucket=bucket, key=key, upload_file=file)

        # Todo add custom error
        except ClientError as err:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def download_objects(self, s3_target: str, local_source: str):
        """
        This method download from s3 target to local data folder
        :param local_source: local data folder i.e. 'D:/Temp/'
        :param s3_target: target s3 path i.e. 'data_store/calc_image_data/'
        :return:
        """
        files = []

        try:
            if '/' in s3_target:
                targets = s3_target.split('/')
                bucket = targets[0]
                key = '/'.join(targets[1:])
            else:
                bucket = s3_target
                key = ''

            response = self.__s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
            if response.get('Contents'):
                for item in response['Contents']:
                    if item['Key'].split('/')[-1]:
                        files.append(item['Key'].split('/')[-1])
                    else:
                        files.append(item['Key'])

            for file in files:
                file_name = os.path.join(local_source, file)
                self.download_file(bucket=bucket, key=key, download_file=file, file_name_path=file_name)

        # Todo add custom error
        except ClientError as err:
            raise
        except Exception:
            raise

    @logger_time_stamp
    @typeguard.typechecked
    def get_last_objects(self, bucket: str, logs_bucket_key: str, policy: str):
        """
        This method return last object per policy, only path without file name
        @param logs_bucket_key:
        @param bucket:
        @param policy:
        @return:
        """
        try:
            if '_' in policy:
                policy = policy.replace('_', '-')
            date_key = (datetime.datetime.now() - datetime.timedelta(days=1)).strftime("%Y/%m/%d")
            objs = self.__s3_client.list_objects_v2(Bucket=bucket,
                                                    Prefix=f'{logs_bucket_key}/{policy}/{date_key}')['Contents']
        except:
            return None
        get_last_modified_key = lambda obj: int(obj['LastModified'].strftime('%s'))
        full_path = [obj['Key'] for obj in sorted(objs, key=get_last_modified_key)][-1]
        return os.path.dirname(full_path)

    @staticmethod
    def __set_default(obj):
        if isinstance(obj, set):
            return list(obj)
        raise TypeError

    @logger_time_stamp
    def save_results_to_s3(self, policy, policy_output, policy_result):
        """
        This method save policy result to s3 with folder creation order by datetime
        @return:
        """

        with gzip.open(self.__resources_file_full_path, 'wt', encoding="ascii") as zipfile:
            json.dump(policy_result, zipfile, default=self.__set_default)
        if 's3' in policy_output:
            s3_operations = S3Operations(self.__region)
            date_key = datetime.datetime.now().strftime("%Y/%m/%d/%H")
            if '/' in policy_output:
                targets = policy_output.split('/')
                bucket = targets[2]
                logs = targets[3]
            s3_operations.upload_file(file_name_path=self.__resources_file_full_path, bucket=bucket,
                                      key=f'{logs}/{self.__region}/{policy}/{date_key}',
                                      upload_file=self.__resource_file_name)
        # save local
        else:
            os.replace(self.__resources_file_full_path, fr'{policy_output}/{self.__resource_file_name}')
        resource_file = self.__resources_file_full_path
        if os.path.isfile(resource_file):
            os.remove(resource_file)

    @logger_time_stamp
    @typeguard.typechecked
    def find_bucket(self, bucket_name: str):
        """
        This method find the bucket is present in the S3 or not
        :param bucket_name:
        :return:
        """
        response = self.__s3_client.list_buckets()
        for bucket in response['Buckets']:
            if bucket['Name'] == bucket_name:
                return True
        return False



import importlib
import inspect

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.ibm.ibm_operations.ibm_operations import IBMOperations


class IBMPolicyRunner(IBMOperations):
    """
    This class run the IBM Policies from policy/ibm
    """

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method run the ibm policies
        @return:
        """
        logger.info(f'account={self._account}, policy={self._policy}, dry_run={self._dry_run}, tag_operation={self._tag_operation}')
        ibm_policy_run = importlib.import_module(f'cloud_governance.policy.ibm.{self._policy}')
        for cls in inspect.getmembers(ibm_policy_run, inspect.isclass):
            if self._policy.replace('_', '') == cls[0].lower():
                response = cls[1]().run()
                if isinstance(response, list) and len(response) > 0:
                    logger.info(f'key: {cls[0]}, count: {len(response)}, {response}')



import os
import smtplib

from cloud_governance.common.logger.init_logger import logger


class Gmail:
    """
    This class send mail by accepting the receivers list and body
    export sender_mail and sender_password
    before creating a password enable two-factor authentication
    create sender_password by GoogleAccount ->  Security --> Apps password
    """

    SMTP_PORT = 465

    def __init__(self):
        self.sender_mail = os.environ.get('SENDER_MAIL', '')
        self.sender_password = os.environ.get('SENDER_PASSWORD', '')

    def send_mail(self, receivers_list: list, subject: str, body: str, cc: list):
        """
        This method sends email
        sender_password: generate from gmail apps
        @param cc:
        @param receivers_list:
        @param subject:
        @param body:
        @return:
        """
        if self.sender_mail and self.sender_password:
            email_text = """\
From: %s
To: %s
Subject: %s
CC: %s

%s
""" % (self.sender_mail, ", ".join(receivers_list), subject, ", ".join(cc), body)
            try:
                smtp_server = smtplib.SMTP_SSL('smtp.gmail.com', self.SMTP_PORT)
                smtp_server.ehlo()
                smtp_server.login(self.sender_mail, self.sender_password)
                smtp_server.sendmail(self.sender_mail, receivers_list, email_text)
                smtp_server.close()
                logger.info(f"""Email sent successfully to the receivers: {receivers_list}!""")
            except Exception as err:
                logger.info(f'Something went wrong…., {err}')
        else:
            logger.info('Sender email and password is requires')



import os
from ast import literal_eval

from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix


class ElasticUpload:

    def __init__(self):
        self.__es_host = os.environ.get('es_host', '')
        self.__es_port = os.environ.get('es_port', '')
        self._es_index = os.environ.get('es_index', '')
        self.account = os.environ.get('account', '').upper()
        self._special_user_mails = os.environ.get('special_user_mails', '{}')
        self._postfix_mail = Postfix()
        self._mail_message = MailMessage()
        if self.__es_host:
            self._elastic_search_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)

    def es_upload_data(self, items: list, es_index: str):
        """
        This method upload data to elastic search
        @param items:
        @param es_index:
        @return:
        """
        try:
            count = 0
            for item in items:
                self._elastic_search_operations.upload_to_elasticsearch(index=es_index, data=item)
                count += 1
            if count > 0 and len(items) > 0:
                logger.info(f'Data Uploaded to {es_index} successfully')
        except Exception as err:
            logger.info(f'Error raised {err}')

    def _literal_eval(self, data: any):
        """
        This method convert string object into its original datatype
        ex: "{'Project': 'Cloud-Governance'}" --> {'Project': 'Cloud-Governance'}
        @param data:
        @return:
        """
        if data:
            return literal_eval(data)
        return data




from cloud_governance.aws.dynamodb_upload_data.upload_data_to_dynamodb import UploadDataToDynamoDb


class CloudTrailToDynamoDb(UploadDataToDynamoDb):
    """
    This class upload cloudtrail data to dynamodb
    """

    def __init__(self):
        super().__init__()

    def __organize_cloudtrail_data(self, cloud_trail_data: list):
        """
        This class filters the cloudtrail data
        @param cloud_trail_data:
        @return:
        """
        organised_data = []
        for data in cloud_trail_data:
            if data.get('Username') not in ('cloud-governance-delete-user', 'cloud-governance-user', 'CloudabilityMonitoring', 'Cloudability') and data.get('EventName') != 'LookupEvents':
                for resource in data.get('Resources'):
                    if resource.get('ResourceName') and not resource.get('ResourceType'):
                        data['ResourceName'] = resource.get('ResourceName')
                    data[resource.get('ResourceType')] = resource.get('ResourceName')
                organised_data.append(data)
        return organised_data

    def __get_cloudtrail_data(self):
        """
        This method extract one day cloudtrail logs
        @return:
        """
        responses = self._cloudtrail_operations.get_regional_cloudtrail_responses(StartTime=self._start_time,
                                                                                  EndTime=self._end_time)
        return responses

    def upload_data(self):
        """
        This method  creates the dynamodb table if not exists and upload the data to table
        @return:
        """
        self._create_table_not_exists(primary_key='EventId')
        cloud_trail_data = self.__get_cloudtrail_data()
        data = self.__organize_cloudtrail_data(cloud_trail_data=cloud_trail_data)
        self._upload_to_dynamo_db_table(data=data)



from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class ZombieSnapshots(NonClusterZombiePolicy):
    """
    This class sends an alert mail for zombie snapshots ( AMI abandoned ) to the user after 4 days and delete after 7 days.
    """

    def __init__(self):
        super().__init__()

    def _get_image_ids_from_description(self, snapshot_description: str):
        """
        This method gets image Ids from snapshot description
        Two cases:
        # Created by CreateImage(i-******) for ami-********
        # Copied for DestinationAmi ami-******* from SourceAmi ami-******* for SourceSnapshot snap-******. Task created on 1,566,308,778,174.
        @return:
        """
        image_ids = []
        images_array = snapshot_description.split('ami-')[1:]
        for image in images_array:
            image_ids.append(f'ami-{image.split(" ")[0]}')
        return image_ids

    def run(self):
        """
        This method return all the zombie snapshots, delete if dry_run no
        @return:
        """
        snapshots = self._ec2_operations.get_snapshots()
        zombie_snapshots = []
        image_ids = self._get_ami_ids()
        for snapshot in snapshots:
            if snapshot.get('Description'):
                snapshot_images = self._get_image_ids_from_description(snapshot.get('Description'))
                tags = snapshot.get('Tags')
                found = False
                for snapshot_image in snapshot_images:
                    if snapshot_image in image_ids:
                        found = True
                snapshot_id = snapshot.get('SnapshotId')
                if not found:
                    unused_days = self._get_resource_last_used_days(tags=tags)
                    zombie_snapshot = self._check_resource_and_delete(resource_name='Snapshot',
                                                                      resource_id='SnapshotId',
                                                                      resource_type='CreateSnapshot',
                                                                      resource=snapshot,
                                                                      empty_days=unused_days,
                                                                      days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE,
                                                                      tags=tags)
                    if zombie_snapshot:
                        zombie_snapshots.append([snapshot.get('SnapshotId'),
                                                 self._get_tag_name_from_tags(tags=tags),
                                                 self._get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                 f'{str(snapshot.get("VolumeSize"))}Gb',
                                                 self._get_policy_value(tags=snapshot.get('Tags')), str(unused_days)
                                                 ])
                else:
                    unused_days = 0
                self._update_resource_tags(resource_id=snapshot_id, tags=tags, left_out_days=unused_days,
                                           resource_left_out=not found)
        return zombie_snapshots



from cloud_governance.common.logger.init_logger import logger
from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class EmptyRoles(NonClusterZombiePolicy):
    """
    This class sends an alert mail for empty role to the user after 4 days and delete after 7 days.
    """

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method return all empty roles, delete if dry_run no
        @return:
        """
        return self.__delete_empty_roles()

    def __delete_empty_roles(self):
        """
        This method deletes the role after 7 days of empty
        @return:
        """
        zombie_roles = []
        roles = self._iam_operations.get_roles()
        for role in roles:
            role_name = role.get('RoleName')
            try:
                get_role = self._iam_client.get_role(RoleName=role.get('RoleName'))['Role']
                tags = get_role.get('Tags')
                if not self._check_live_cluster_tag(tags=tags):
                    role_empty = False
                    role_attached_policies = self._iam_client.list_attached_role_policies(RoleName=role_name)
                    role_inline_policies = self._iam_client.list_role_policies(RoleName=role_name)
                    if not role_inline_policies.get('PolicyNames') and not role_attached_policies.get('AttachedPolicies'):
                        role_empty = True
                        if not self._get_tag_name_from_tags(tags=tags, tag_name='Name'):
                            tags.append({'Key': 'Name', 'Value': role_name})
                        empty_days = self._get_resource_last_used_days(tags=tags)
                        empty_role = self._check_resource_and_delete(resource_name='IAM Role', resource_id='RoleName', resource_type='CreateRole', resource=get_role, empty_days=empty_days, days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE, tags=tags)
                        if empty_role:
                            zombie_roles.append([empty_role.get('RoleName'), self._get_tag_name_from_tags(tags=tags, tag_name='User'), self._get_policy_value(tags=tags), empty_days])
                    else:
                        empty_days = 0
                    self._update_resource_tags(resource_id=role_name, tags=tags, left_out_days=empty_days, resource_left_out=role_empty)
            except Exception as err:
                logger.info(f'Error occur:{role_name}, {err}')
        return zombie_roles



import os
from ast import literal_eval

from cloud_governance.common.clouds.ibm.account.ibm_account import IBMAccount
from cloud_governance.common.clouds.ibm.classic.classic_operations import ClassicOperations


class TaggingOperations:
    """
    This class contain common methods and parameters required for tagging ibm baremetal, vm
    """

    def __init__(self):
        self._ibm_client = IBMAccount()
        self._sl_client = self._ibm_client.get_sl_client()
        self._classic_operations = ClassicOperations()
        self._dry_run = os.environ.get('dry_run', 'yes')
        self._tag_operation = os.environ.get('tag_operation', 'read')
        self._tag_remove_name = os.environ.get('tag_remove_name', '')
        self._tag_custom = self.__get_literal_eval(os.environ.get('tag_custom', '[]'))

    def __get_literal_eval(self, env_input: any):
        return literal_eval(env_input)

    def _filter_common_tags(self, user_tags: list, resource_tags: list):
        """
        This method filter the common tags form the resource tags to user tags
        @param user_tags:
        @param resource_tags:
        @return:
        """
        tags = []
        for tag in user_tags:
            if tag not in resource_tags:
                tags.append(tag)
        return tags

    def _filter_remove_tags(self, user_tags: list, resource_tags: list):
        """
        This method filter the remove tags from use_tags to resource_tags
        @param user_tags:
        @param resource_tags:
        @return:
        """
        remove_tags = []
        for tag in user_tags:
            if tag in resource_tags:
                remove_tags.append(tag)
        return remove_tags

    def softlayer_operation(self, softlayer_name: str, softlayer_method: str, resource_id: str, tags: str):
        """
        This method takes softlayer name and its method
        @param softlayer_name:
        @param softlayer_method:
        @param resource_id:
        @param tags:
        @return:
        """
        return self._sl_client.call(softlayer_name,softlayer_method, tags, id=resource_id)



import datetime
import os

import pandas
from typeguard import typechecked

from cloud_governance.common.clouds.ibm.account.ibm_account import IBMAccount
from cloud_governance.common.clouds.ibm.classic.classic_operations import ClassicOperations
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload


class IBMCostReport(ElasticUpload):
    """
    This class fetched the invoice reports from the IBM based on the monthly and upload to ElasticSearch.
    """

    def __init__(self):
        super().__init__()
        self.ibm_account = IBMAccount()
        self.classic_operations = ClassicOperations()
        self.month = os.environ.get('month', '')
        self.year = os.environ.get('year', '')
        self.owned_tags = ['owner', 'budget', 'environment', 'manager', 'user', 'project', 'fqdn']

    @typechecked
    def collect_tags_from_machines(self, tags: list):
        """
        This method return tags from list of string tags
        @param tags:
        @return:
        """
        hardware_tags = {}
        if tags:
            for tag in tags:
                if ':' in tag:
                    key, value = tag.split(':')
                    if key in self.owned_tags:
                        hardware_tags[key] = value
        return hardware_tags

    def get_hardware_data(self):
        """
        This method returns the baremetal resource data
        @return:
        """
        bare_metals = self.classic_operations.get_hardware_ids()
        collect_machines_data = {}
        for hardware in bare_metals:
            hardware_tags = self.collect_tags_from_machines(
                tags=self.classic_operations.get_hardware_tags(hardware_id=str(hardware.get('id'))))
            hardware_tags['fqdn'] = hardware.get('fullyQualifiedDomainName').lower()
            collect_machines_data[hardware_tags['fqdn']] = hardware_tags
        return collect_machines_data

    def get_virtual_machine_data(self):
        """
        This method returns the virtual machine data
        @return:
        """
        vms = self.classic_operations.get_virtual_machine_ids()
        collect_machines_data = {}
        for vm in vms:
            vm_tags = self.collect_tags_from_machines(
                tags=self.classic_operations.get_virtual_machine_tags(vm_id=str(vm.get('id'))))
            vm_tags['fqdn'] = vm.get('fullyQualifiedDomainName').lower()
            collect_machines_data[vm_tags['fqdn']] = vm_tags
        return collect_machines_data

    @typechecked
    def get_invoice_data(self, collect_machines_data: dict):
        """
        This method returns the invoice data
        @return:
        """
        month, year = self.month, self.year
        if not month and not year:
            date = datetime.datetime.now().date()
            month, year = date.month, date.year
        else:
            month = int(month)
            year = int(year)
        invoice_data, users = self.ibm_account.get_invoice_data(month=month, year=year)
        for fqdn in invoice_data.keys():
            if fqdn not in collect_machines_data:
                user = [user for user in users if user in fqdn]
                if user:
                    user_tags = self.ibm_account.get_user_tags_from_gsheet(username=f'{user[0]}@redhat.com', user_email='yes')
                    invoice_data[fqdn].update(self.collect_tags_from_machines(user_tags))
        return invoice_data

    @typechecked
    def concatenate_dictionaries(self, resource_data: dict, invoice_data: dict):
        """
        This method concatenate two dicts and aggregate its sum
        @param resource_data:
        @param invoice_data:
        @return:
        """
        resource_df = pandas.DataFrame(list(resource_data.values()))
        invoice_df = pandas.DataFrame(list(invoice_data.values()))
        concat_data_df = pandas.concat([resource_df, invoice_df], ignore_index=True)
        concat_data_df = concat_data_df.groupby('fqdn').agg('sum', '').reset_index()
        return concat_data_df.to_dict(orient='index')

    def run(self):
        """
        This method upload the invoice cost report by aggregating tags to elasticsearch
        @return:
        """
        collect_machines_data = self.get_hardware_data()
        collect_machines_data.update(self.get_virtual_machine_data())
        invoice_data = self.get_invoice_data(collect_machines_data=collect_machines_data)
        cost_invoice_resource_data = self.concatenate_dictionaries(resource_data=collect_machines_data, invoice_data=invoice_data)
        for tag_name in self.owned_tags:
            cost_list_items = []
            for _, data in cost_invoice_resource_data.items():
                if data[tag_name] != 0 or tag_name == 'budget':
                    cost_list_items.append({
                        tag_name.capitalize(): data[tag_name],
                        'Cost': data['cost'],
                        'Budget': self.account
                    })
            self.es_upload_data(items=cost_list_items, es_index=f'{self._es_index}-{tag_name.lower()}')
        return True



from datetime import datetime

import boto3

from cloud_governance.common.clouds.aws.cloudtrail.cloudtrail_operations import CloudTrailOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.clouds.aws.utils.utils import Utils


class NonClusterOperations:

    def __init__(self, region: str = 'us-east-2', dry_run: str = 'yes', input_tags: dict = ''):
        self.region = region
        self.dry_run = dry_run
        self.input_tags = input_tags
        self.cloudtrail = boto3.client('cloudtrail', region_name=region)
        self.cluster_prefix = 'kubernetes.io/cluster/'
        self.ec2_client = boto3.client('ec2', region_name=region)
        self.cloudtrail = CloudTrailOperations(region_name=self.region)
        self.iam_client = IAMOperations()
        self.ec2_operations = EC2Operations(region=region)
        self.utils = Utils(region=region)

    def _get_instances_data(self, instance_id: str = ''):
        """
        This method go over all instances
        :return:
        """
        ec2s_data = self.ec2_operations.get_instances()
        if instance_id:
            for items in ec2s_data:
                if items.get('Instances'):
                    for item in items['Instances']:
                        if item.get('InstanceId') == instance_id:
                            return items['Instances']
        else:
            instances_list = []
            for items in ec2s_data:
                if items.get('Instances'):
                    instances_list.append(items['Instances'])
            return instances_list

    def _append_input_tags(self):
        """
        This method build tags list according to input tags dictionary
        @return:
        """
        tags_list = []
        for key, value in self.input_tags.items():
            tags_list.append({'Key': key, 'Value': value})
        return tags_list

    def _get_tags_of_resources(self, tags: list, search_tags: list):
        """
        This method extracts tags from the resource tags
        @param tags:
        @param search_tags:
        @return:
        """
        add_tags = []
        if tags:
            for search_tag in search_tags:
                found = False
                for tag in tags:
                    if tag.get('Key') == search_tag.get('Key'):
                        found = True
                if not found:
                    add_tags.append(search_tag)
        else:
            add_tags.extend(search_tags)
        return add_tags

    def _fill_na_tags(self, user: str = None):
        """
        This method fill NA tags
        @param user:
        @return:
        """
        tags = []
        keys = ['User', 'Owner', 'Manager', 'Project', 'Environment', 'Email']
        value = 'NA'
        for key in keys:
            if user and key == 'User':
                tags.append({'Key': key, 'Value': user})
            elif user and key == 'Email':
                tags.append({'Key': key, 'Value': f'{user}@redhat.com'})
            else:
                tags.append({'Key': key, 'Value': value})
        return tags

    def _get_username_from_cloudtrail(self, start_time: datetime, resource_id: str, resource_type: str):
        """
        This method return username fom cloudtrail
        @param start_time:
        @param resource_id:
        @param resource_type:
        @return:
        """
        return self.cloudtrail.get_username_by_instance_id_and_time(start_time=start_time, resource_id=resource_id, resource_type=resource_type)

    def _get_resource_data(self, resource_method: callable):
        """
        This method returns the resource_data
        @return:
        """
        resource_data = resource_method()
        if 'ec2_operations' not in str(resource_method):
            _, resource_list = self.ec2_operations.scan_cluster_or_non_cluster_instance(resource_data)
        else:
            _, resource_list = self.ec2_operations.scan_cluster_non_cluster_resources(resource_data)
        return resource_list

    def _convert_datetime_format(self, date_time: datetime):
        """
        This method convert datetime to YYYY/MM/DD H:M:S
        @param date_time:
        @return:
        """
        return date_time.strftime('%Y/%m/%d %H:%M:%S')

    def _build_tag(self, key: str, value: any):
        """
        This method return Key value pair
        @param key:
        @param value:
        @return:
        """
        if isinstance(value, datetime):
            value = self._convert_datetime_format(date_time=value)
        return {'Key': key, 'Value': value}

    def _get_tags_from_instance_item(self, instance_item: dict):
        tags = []
        username = ''
        if instance_item.get('Tags'):
            tags.extend(
                [tag for tag in instance_item.get('Tags') if not tag.get('Key') == 'Name'])
            for tag in instance_item.get('Tags'):
                if tag.get('Key') == 'User':
                    username = tag.get('Value')
                elif tag.get('Key') == 'LaunchTime':
                    tags.append({'Key': 'LaunchTime', 'Value': tag.get('Key')})
        if not username:
            tags.extend(self._append_input_tags())
            username = self._get_username_from_cloudtrail(start_time=instance_item.get('LaunchTime'),
                                                          resource_id=instance_item.get('InstanceId'),
                                                          resource_type='AWS::EC2::Instance')
            launch_time = instance_item.get('LaunchTime')
            tags.append(self._build_tag(key='LaunchTime', value=launch_time))
        return tags, username

    def _get_tags_fom_attachments(self, attachments: list):
        """
        This method return tags from attachments
        @param attachments:
        @return:
        """
        tags = []
        username = ''
        if attachments:
            for attachment in attachments:
                for item in self._get_instances_data(attachment.get('InstanceId')):
                    item_tags, username = self._get_tags_from_instance_item(instance_item=item)
        else:
            tags.extend(self._append_input_tags())
        return tags, username

    def _get_tags_from_snapshot_description_images(self, description: str):
        tags = []
        username = ''
        if description and 'Created' in description:
            image_id = description.split(" ")[-1]
            images = self.ec2_client.describe_images(Owners=['self'])['Images']
            for image in images:
                if image.get('ImageId') == image_id:
                    if image.get('tags'):
                        tags.extend(
                            [tag for tag in image.get('Tags') if not tag.get('Key') == "Name"])
                    else:
                        tags.extend(self._append_input_tags())
                    start_time = datetime.fromisoformat(image.get('CreationDate')[:-1] + '+00:00')
                    username = self._get_username_from_cloudtrail(start_time=start_time, resource_id=image_id, resource_type='AWS::EC2::Ami')
        return tags, username


